# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.block_storage.v3 import group_type as _group_type
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _format_group_type(
    group_type: _group_type.GroupType,
) -> tuple[Sequence[str], Iterable[Any]]:
    columns = (
        'id',
        'name',
        'description',
        'is_public',
        'group_specs',
    )
    column_headers = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )

    return (
        column_headers,
        utils.get_item_properties(
            group_type,
            columns,
            formatters={'group_specs': format_columns.DictColumn},
        ),
    )


class CreateVolumeGroupType(command.ShowOne):
    """Create a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of new volume group type.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the volume group type.'),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--public',
            dest='is_public',
            action='store_true',
            default=True,
            help=_(
                'Volume group type is available to other projects (default)'
            ),
        )
        type_group.add_argument(
            '--private',
            dest='is_public',
            action='store_false',
            help=_('Volume group type is not available to other projects'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type create' command"
            )
            raise exceptions.CommandError(msg)

        group_type = volume_client.create_group_type(
            name=parsed_args.name,
            description=parsed_args.description,
            is_public=parsed_args.is_public,
        )

        return _format_group_type(group_type)


class DeleteVolumeGroupType(command.Command):
    """Delete a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type delete' command"
            )
            raise exceptions.CommandError(msg)

        group_type = volume_client.find_group_type(
            parsed_args.group_type,
            ignore_missing=False,
        )

        volume_client.delete_group_type(group_type)


class SetVolumeGroupType(command.ShowOne):
    """Update a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name for volume group type.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New description for volume group type.'),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--public',
            dest='is_public',
            action='store_true',
            default=None,
            help=_('Make volume group type available to other projects.'),
        )
        type_group.add_argument(
            '--private',
            dest='is_public',
            action='store_false',
            help=_('Make volume group type unavailable to other projects.'),
        )
        parser.add_argument(
            '--no-property',
            action='store_true',
            help=_(
                'Remove all properties from this volume group type '
                '(specify both --no-property and --property '
                'to remove the current properties before setting '
                'new properties)'
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Property to add or modify for this volume group type '
                '(repeat option to set multiple properties)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type set' command"
            )
            raise exceptions.CommandError(msg)

        group_type = volume_client.find_group_type(
            parsed_args.group_type,
            ignore_missing=False,
        )

        kwargs = {}
        errors = 0

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name

        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description

        if parsed_args.is_public is not None:
            kwargs['is_public'] = parsed_args.is_public

        if kwargs:
            try:
                group_type = volume_client.update_group_type(
                    group_type.id, **kwargs
                )
            except Exception as e:
                LOG.error(_("Failed to update group type: %s"), e)
                errors += 1

        if parsed_args.no_property:
            try:
                for key in group_type.group_specs:
                    volume_client.delete_group_type_group_specs_property(
                        group_type.id, key
                    )
            except Exception as e:
                LOG.error(_("Failed to clear group type properties: %s"), e)
                errors += 1

        if parsed_args.properties:
            try:
                volume_client.create_group_type_group_specs(
                    group_type.id, parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to set group type properties: %s"), e)
                errors += 1

        if errors > 0:
            msg = _("Command Failed: One or more of the operations failed")
            raise exceptions.CommandError(msg)

        group_type = volume_client.find_group_type(
            parsed_args.group_type,
            ignore_missing=False,
        )

        return _format_group_type(group_type)


class UnsetVolumeGroupType(command.ShowOne):
    """Unset properties of a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Property to remove from this volume group type '
                '(repeat option to unset multiple properties)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type unset' command"
            )
            raise exceptions.CommandError(msg)

        for key in parsed_args.properties or []:
            volume_client.delete_group_type_group_specs_property(
                parsed_args.group_type, key
            )

        group_type = volume_client.find_group_type(
            parsed_args.group_type,
            ignore_missing=False,
        )

        return _format_group_type(group_type)


class ListVolumeGroupType(command.Lister):
    """Lists all volume group types.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--default',
            action='store_true',
            dest='show_default',
            default=False,
            help=_('List the default volume group type.'),
        )
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type list' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.show_default:
            group_types = [volume_client.get_group_type('default')]
        else:
            group_types = list(volume_client.group_types())

        column_headers = (
            'ID',
            'Name',
            'Is Public',
            'Group Specs',
        )
        columns = (
            'ID',
            'Name',
            'Is Public',
            'Group Specs',
        )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    a,
                    columns,
                    formatters={'Group Specs': format_columns.DictColumn},
                )
                for a in group_types
            ),
        )


class ShowVolumeGroupType(command.ShowOne):
    """Show detailed information for a volume group type.

    This command requires ``--os-volume-api-version`` 3.11 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group_type',
            metavar='<group_type>',
            help=_('Name or ID of volume group type.'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.11'):
            msg = _(
                "--os-volume-api-version 3.11 or greater is required to "
                "support the 'volume group type show' command"
            )
            raise exceptions.CommandError(msg)

        group_type = volume_client.find_group_type(
            parsed_args.group_type,
            ignore_missing=False,
        )

        return _format_group_type(group_type)



# Copyright (c) 2016 Juniper Networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


from openstack.network.v2 import bgpvpn as _bgpvpn
from openstack.network.v2 import bgpvpn_network_association as _net_assoc
from openstack.network.v2 import bgpvpn_port_association as _port_assoc
from openstack.network.v2 import bgpvpn_router_association as _router_assoc
from openstack import resource as sdk_resource

from openstackclient.tests.unit.network.v2 import fakes as test_fakes


_FAKE_PROJECT_ID = 'fake_project_id'


class TestNeutronClientBgpvpn(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()


def create_one_bgpvpn(attrs=None):
    """Create a fake BGP VPN."""

    attrs = attrs or {}

    # Set default attributes.
    bgpvpn_attrs = {
        'id': 'fake_bgpvpn_id',
        'project_id': _FAKE_PROJECT_ID,
        'name': '',
        'type': 'l3',
        'route_targets': [],
        'import_targets': [],
        'export_targets': [],
        'route_distinguishers': [],
        'networks': [],
        'routers': [],
        'ports': [],
        'vni': 100,
        'local_pref': 777,
    }

    # Overwrite default attributes.
    bgpvpn_attrs.update(attrs)
    return _bgpvpn.BgpVpn(**bgpvpn_attrs)


def create_bgpvpns(attrs=None, count=1):
    """Create multiple fake BGP VPN."""

    bgpvpns = []
    for i in range(0, count):
        if attrs is None:
            attrs = {'id': 'fake_id{i}'}
        elif getattr(attrs, 'id', None) is None:
            attrs['id'] = 'fake_id{i}'
        bgpvpns.append(create_one_bgpvpn(attrs))

    return bgpvpns


def create_one_resource(attrs=None):
    """Create a fake resource."""
    attrs = attrs or {}

    res_attrs = {
        'id': 'fake_resource_id',
    }

    res_attrs.update(attrs)
    return sdk_resource.Resource(**res_attrs)


def create_resources(attrs=None, count=1):
    """Create multiple fake resources."""

    resources = []
    for i in range(0, count):
        if attrs is None:
            attrs = {'id': 'fake_id{i}'}
        elif getattr(attrs, 'id', None) is None:
            attrs['id'] = 'fake_id{i}'
        resources.append(create_one_resource(attrs))

    return resources


def create_one_network_association(attrs=None):
    """Create a fake network association."""
    attrs = attrs or {}
    assoc_attrs = {
        'id': 'fake_association_id',
        'network_id': 'fake_resource_id',
        'project_id': _FAKE_PROJECT_ID,
    }
    assoc_attrs.update(attrs)
    return _net_assoc.BgpVpnNetworkAssociation(**assoc_attrs)


def create_network_associations(count=1):
    """Create multiple fake network associations."""
    assocs = []
    for idx in range(count):
        assocs.append(
            {
                'id': f'fake_association_id{idx}',
                'network_id': f'fake_resource_id{idx}',
                'project_id': _FAKE_PROJECT_ID,
            }
        )
    return assocs


def create_one_router_association(attrs=None):
    """Create a fake router association."""
    attrs = attrs or {}
    assoc_attrs = {
        'id': 'fake_association_id',
        'router_id': 'fake_resource_id',
        'project_id': _FAKE_PROJECT_ID,
    }
    assoc_attrs.update(attrs)
    return _router_assoc.BgpVpnRouterAssociation(**assoc_attrs)


def create_router_associations(count=1):
    """Create multiple fake router associations."""
    assocs = []
    for idx in range(count):
        assocs.append(
            {
                'id': f'fake_association_id{idx}',
                'router_id': f'fake_resource_id{idx}',
                'project_id': _FAKE_PROJECT_ID,
            }
        )
    return assocs


def create_one_port_association(attrs=None):
    """Create a fake port association."""
    attrs = attrs or {}
    assoc_attrs = {
        'id': 'fake_association_id',
        'port_id': 'fake_resource_id',
        'project_id': _FAKE_PROJECT_ID,
        'routes': [],
    }
    assoc_attrs.update(attrs)
    return _port_assoc.BgpVpnPortAssociation(**assoc_attrs)


def create_port_associations(count=1):
    """Create multiple fake port associations."""
    assocs = []
    for idx in range(count):
        assocs.append(
            {
                'id': f'fake_association_id{idx}',
                'port_id': f'fake_resource_id{idx}',
                'project_id': _FAKE_PROJECT_ID,
                'routes': [],
            }
        )
    return assocs



#   Copyright 2018 SUSE Linux GmbH
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Application Credential action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import datetime
import json
import logging
from typing import Any
import uuid

from cliff import columns as cliff_columns
from openstack.identity.v3 import (
    application_credential as _application_credential,
)
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common

LOG = logging.getLogger(__name__)


class RolesColumn(cliff_columns.FormattableColumn[Any]):
    """Generate a formatted string of role names."""

    def human_readable(self) -> str:
        return utils.format_list(list(r['name'] for r in self._value)) or ""


def _format_application_credential(
    application_credential: _application_credential.ApplicationCredential,
    *,
    include_secret: bool = False,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    column_headers: tuple[str, ...] = (
        'ID',
        'Name',
        'Description',
        'Project ID',
        'Roles',
        'Unrestricted',
        'Access Rules',
        'Expires At',
    )
    columns: tuple[str, ...] = (
        'id',
        'name',
        'description',
        'project_id',
        'roles',
        'unrestricted',
        'access_rules',
        'expires_at',
    )
    if include_secret:
        column_headers += ('Secret',)
        columns += ('secret',)

    return (
        column_headers,
        utils.get_item_properties(
            application_credential, columns, formatters={'roles': RolesColumn}
        ),
    )


def _format_application_credentials(
    application_credentials: Iterable[
        _application_credential.ApplicationCredential
    ],
) -> tuple[tuple[str, ...], Any]:
    column_headers = (
        'ID',
        'Name',
        'Description',
        'Project ID',
        'Roles',
        'Unrestricted',
        'Access Rules',
        'Expires At',
    )
    columns = (
        'id',
        'name',
        'description',
        'project_id',
        'roles',
        'unrestricted',
        'access_rules',
        'expires_at',
    )

    return (
        column_headers,
        (
            utils.get_item_properties(
                x, columns, formatters={'roles': RolesColumn}
            )
            for x in application_credentials
        ),
    )


# TODO(stephenfin): Move this to osc_lib since it's useful elsewhere
def is_uuid_like(value: str) -> bool:
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool
    """
    try:
        formatted_value = (
            value.replace('urn:', '')
            .replace('uuid:', '')
            .strip('{}')
            .replace('-', '')
            .lower()
        )
        return str(uuid.UUID(value)).replace('-', '') == formatted_value
    except (TypeError, ValueError, AttributeError):
        return False


class CreateApplicationCredential(command.ShowOne):
    _description = _("Create new application credential")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of the application credential'),
        )
        parser.add_argument(
            '--secret',
            metavar='<secret>',
            help=_(
                'Secret to use for authentication (if not provided, one'
                ' will be generated)'
            ),
        )
        parser.add_argument(
            '--role',
            metavar='<role>',
            dest='roles',
            action='append',
            default=[],
            help=_(
                'Roles to authorize (name or ID) (repeat option to set'
                ' multiple values)'
            ),
        )
        parser.add_argument(
            '--expiration',
            metavar='<expiration>',
            help=_(
                'Sets an expiration date for the application credential,'
                ' format of YYYY-mm-ddTHH:MM:SS (if not provided, the'
                ' application credential will not expire)'
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Application credential description'),
        )
        parser.add_argument(
            '--unrestricted',
            action="store_true",
            help=_(
                'Enable application credential to create and delete other'
                ' application credentials and trusts (this is potentially'
                ' dangerous behavior and is disabled by default)'
            ),
        )
        parser.add_argument(
            '--restricted',
            action="store_true",
            help=_(
                'Prohibit application credential from creating and deleting'
                ' other application credentials and trusts (this is the'
                ' default behavior)'
            ),
        )
        parser.add_argument(
            '--access-rules',
            metavar='<access-rules>',
            help=_(
                'Either a string or file path containing a JSON-formatted '
                'list of access rules, each containing a request method, '
                'path, and service, for example '
                '\'[{"method": "GET", '
                '"path": "/v2.1/servers", '
                '"service": "compute"}]\''
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')

        user_id = auth.get_user_id(conn.session)
        if user_id is None:
            msg = _("failed to retrieve auth info for current session")
            raise exceptions.CommandError(msg)

        role_ids = []
        for role in parsed_args.roles:
            if is_uuid_like(role):
                role_ids.append({'id': role})
            else:
                role_ids.append({'name': role})

        expires_at = None
        if parsed_args.expiration:
            expires_at = datetime.datetime.strptime(
                parsed_args.expiration, '%Y-%m-%dT%H:%M:%S'
            )

        if parsed_args.restricted:
            unrestricted = False
        else:
            unrestricted = parsed_args.unrestricted

        if parsed_args.access_rules:
            try:
                access_rules = json.loads(parsed_args.access_rules)
            except ValueError:
                try:
                    with open(parsed_args.access_rules) as f:
                        access_rules = json.load(f)
                except OSError:
                    msg = _(
                        "Access rules is not valid JSON string or file does"
                        " not exist."
                    )
                    raise exceptions.CommandError(msg)
        else:
            access_rules = []

        application_credential = identity_client.create_application_credential(
            user_id,
            parsed_args.name,
            roles=role_ids,
            expires_at=expires_at,
            description=parsed_args.description,
            secret=parsed_args.secret,
            unrestricted=unrestricted,
            access_rules=access_rules,
        )

        return _format_application_credential(
            application_credential, include_secret=True
        )


class DeleteApplicationCredential(command.Command):
    _description = _("Delete application credentials(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'application_credential',
            metavar='<application-credential>',
            nargs="+",
            help=_('Application credentials(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')

        user_id = auth.get_user_id(conn.session)
        if user_id is None:
            msg = _("failed to retrieve auth info for current session")
            raise exceptions.CommandError(msg)

        errors = 0
        for ac in parsed_args.application_credential:
            try:
                app_cred = identity_client.find_application_credential(
                    user_id, ac, ignore_missing=False
                )
                identity_client.delete_application_credential(
                    user_id, app_cred.id
                )
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete application credential with "
                        "name or ID '%(ac)s': %(e)s"
                    ),
                    {'ac': ac, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.application_credential)
            msg = _(
                "%(errors)s of %(total)s application credentials failed "
                "to delete."
            ) % {'errors': errors, 'total': total}
            raise exceptions.CommandError(msg)

        return None


class ListApplicationCredential(command.Lister):
    _description = _("List application credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('User whose application credentials to list (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        if parsed_args.user:
            user_id = common.find_user_id_sdk(
                identity_client, parsed_args.user, parsed_args.user_domain
            )
        else:
            conn = self.app.client_manager.sdk_connection
            auth = conn.config.get_auth()
            if auth is None:
                # this will never happen
                raise exceptions.CommandError('invalid authentication info')
            _user_id = auth.get_user_id(conn.session)
            if _user_id is None:
                # this will never happen
                raise exceptions.CommandError('invalid authentication info')
            user_id = _user_id

        application_credentials = identity_client.application_credentials(
            user=user_id
        )

        return _format_application_credentials(application_credentials)


class ShowApplicationCredential(command.ShowOne):
    _description = _("Display application credential details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'application_credential',
            metavar='<application-credential>',
            help=_('Application credential to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')

        user_id = auth.get_user_id(conn.session)
        if user_id is None:
            msg = _("failed to retrieve auth info for current session")
            raise exceptions.CommandError(msg)

        application_credential = identity_client.find_application_credential(
            user_id, parsed_args.application_credential, ignore_missing=False
        )

        return _format_application_credential(application_credential)



#    Copyright 2017 FUJITSU LIMITED
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network import v2 as network_v2
from osc_lib.cli import identity as identity_utils
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


_attr_map = [
    ('id', 'ID', column_util.LIST_BOTH),
    ('name', 'Name', column_util.LIST_BOTH),
    ('router_id', 'Router', column_util.LIST_BOTH),
    ('subnet_id', 'Subnet', column_util.LIST_BOTH),
    ('flavor_id', 'Flavor', column_util.LIST_BOTH),
    ('is_admin_state_up', 'State', column_util.LIST_BOTH),
    ('status', 'Status', column_util.LIST_BOTH),
    ('description', 'Description', column_util.LIST_LONG_ONLY),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
    ('external_v4_ip', 'Ext v4 IP', column_util.LIST_LONG_ONLY),
    ('external_v6_ip', 'Ext v6 IP', column_util.LIST_LONG_ONLY),
]

_attr_map_dict = {
    'id': 'ID',
    'name': 'Name',
    'router_id': 'Router',
    'subnet_id': 'Subnet',
    'flavor_id': 'Flavor',
    'is_admin_state_up': 'State',
    'status': 'Status',
    'description': 'Description',
    'project_id': 'Project',
    'external_v4_ip': 'Ext v4 IP',
    'external_v6_ip': 'Ext v6 IP',
}


def _get_common_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        '--description',
        metavar='<description>',
        help=_('Description for the VPN service'),
    )
    parser.add_argument(
        '--subnet',
        metavar='<subnet>',
        help=_('Local private subnet (name or ID)'),
    )
    parser.add_argument(
        '--flavor',
        metavar='<flavor>',
        help=_('Flavor for the VPN service (name or ID)'),
    )
    admin_group = parser.add_mutually_exclusive_group()
    admin_group.add_argument(
        '--enable', action='store_true', help=_("Enable VPN service")
    )
    admin_group.add_argument(
        '--disable', action='store_true', help=_("Disable VPN service")
    )
    return parser


def _get_common_attrs(
    client: network_v2.Proxy, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    if parsed_args.description:
        attrs['description'] = str(parsed_args.description)
    if parsed_args.subnet:
        _subnet_id = client.find_subnet(
            parsed_args.subnet, ignore_missing=False
        ).id
        attrs['subnet_id'] = _subnet_id
    if parsed_args.flavor:
        _flavor_id = client.find_flavor(
            parsed_args.flavor, ignore_missing=False
        ).id
        attrs['flavor_id'] = _flavor_id
    if parsed_args.enable:
        attrs['admin_state_up'] = True
    if parsed_args.disable:
        attrs['admin_state_up'] = False
    return attrs


class CreateVPNService(command.ShowOne):
    _description = _("Create an VPN service")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        parser.add_argument(
            'name', metavar='<name>', help=_('Name for the VPN service')
        )
        parser.add_argument(
            '--router',
            metavar='ROUTER',
            required=True,
            help=_('Router for the VPN service (name or ID)'),
        )
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_common_attrs(client, parsed_args)
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id
        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        if parsed_args.router:
            _router_id = client.find_router(
                parsed_args.router, ignore_missing=False
            ).id
            attrs['router_id'] = _router_id
        obj = client.create_vpn_service(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteVPNService(command.Command):
    _description = _("Delete VPN service(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'vpnservice',
            metavar='<vpn-service>',
            nargs='+',
            help=_('VPN service to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0
        for vpn in parsed_args.vpnservice:
            try:
                vpn_id = client.find_vpn_service(vpn, ignore_missing=False).id
                client.delete_vpn_service(vpn_id)
            except Exception as e:
                result += 1
                print(
                    f"Failed to delete VPN service with name or ID {vpn}: {e}"
                )

        if result > 0:
            total = len(parsed_args.vpnservice)
            msg = _(
                "%(result)s of %(total)s vpn service failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListVPNService(command.Lister):
    _description = _("List VPN services that belong to a given project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        obj = client.vpn_services()
        headers, columns = column_util.get_column_definitions(
            _attr_map, long_listing=parsed_args.long
        )
        return (headers, (utils.get_dict_properties(s, columns) for s in obj))


class SetVPNSercice(command.Command):
    _description = _("Set VPN service properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        _get_common_parser(parser)
        parser.add_argument(
            '--name', metavar='<name>', help=_('Name for the VPN service')
        )
        parser.add_argument(
            'vpnservice',
            metavar='<vpn-service>',
            help=_('VPN service to modify (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        attrs = _get_common_attrs(client, parsed_args)
        if parsed_args.name:
            attrs['name'] = str(parsed_args.name)
        vpn_id = client.find_vpn_service(
            parsed_args.vpnservice, ignore_missing=False
        )['id']
        try:
            client.update_vpn_service(vpn_id, **attrs)
        except Exception as e:
            msg = _("Failed to set vpn service '%(vpn)s': %(e)s") % {
                'vpn': parsed_args.vpnservice,
                'e': e,
            }
            raise exceptions.CommandError(msg)


class ShowVPNService(command.ShowOne):
    _description = _("Display VPN service details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'vpnservice',
            metavar='<vpn-service>',
            help=_('VPN service to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_vpn_service(
            parsed_args.vpnservice, ignore_missing=False
        )
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return (display_columns, data)



#   Copyright 2017 FUJITSU LIMITED
#   All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from openstack.identity.v3 import project as _project
from openstack.network.v2 import vpn_endpoint_group as _vpn_endpoint_group
from openstack.test import fakes as sdk_fakes
from openstackclient.tests.unit import utils as tests_utils
from osc_lib import exceptions

from openstackclient.network.v2.vpnaas import endpoint_group
from openstackclient.tests.unit.network.v2 import fakes as test_fakes
from openstackclient.tests.unit.network.v2.vpnaas import fakes


class TestEndpointGroup(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self._endpoint_group = fakes.EndpointGroup().create()

        def _mock_endpoint_group(*args, **kwargs):
            self.network_client.find_vpn_endpoint_group.assert_called_once_with(
                self._endpoint_group['id'], ignore_missing=False
            )
            return {'id': args[0]}

        self.network_client.find_vpn_endpoint_group.side_effect = mock.Mock(
            side_effect=_mock_endpoint_group
        )
        self.headers = (
            'ID',
            'Name',
            'Type',
            'Endpoints',
            'Description',
            'Project',
        )
        self.ordered_headers = (
            'Description',
            'Endpoints',
            'ID',
            'Name',
            'Project',
            'Type',
        )
        self.ordered_data = (
            self._endpoint_group['description'],
            self._endpoint_group['endpoints'],
            self._endpoint_group['id'],
            self._endpoint_group['name'],
            self._endpoint_group['project_id'],
            self._endpoint_group['type'],
        )
        self.ordered_columns = (
            'description',
            'endpoints',
            'id',
            'name',
            'project_id',
            'type',
        )


class TestCreateEndpointGroup(TestEndpointGroup):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.network_client.create_vpn_endpoint_group.return_value = (
            self._endpoint_group
        )

        self.cmd = endpoint_group.CreateEndpointGroup(self.app, None)

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_all_params_cidr(self):
        self.network_client.create_vpn_endpoint_group.return_value = (
            self._endpoint_group
        )

        arglist = [
            '--description',
            'my-desc',
            '--type',
            'cidr',
            '--value',
            '10.0.0.0/24',
            '--value',
            '20.0.0.0/24',
            '--project',
            self.project.id,
            self._endpoint_group.name,
        ]
        verifylist = [
            ('description', 'my-desc'),
            ('type', 'cidr'),
            ('endpoints', ['10.0.0.0/24', '20.0.0.0/24']),
            ('project', self.project.id),
            ('name', self._endpoint_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.ordered_data = tuple(
            self._endpoint_group[column] for column in self.ordered_columns
        )
        self.network_client.create_vpn_endpoint_group.assert_called_once_with(
            description='my-desc',
            type='cidr',
            endpoints=['10.0.0.0/24', '20.0.0.0/24'],
            project_id=self.project.id,
            name=self._endpoint_group.name,
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertEqual(self.ordered_data, data)


class TestDeleteEndpointGroup(TestEndpointGroup):
    def setUp(self):
        super().setUp()
        self.cmd = endpoint_group.DeleteEndpointGroup(self.app, None)

    def test_delete_with_one_resource(self):
        self.network_client.find_vpn_endpoint_group.side_effect = [
            self._endpoint_group
        ]

        arglist = [self._endpoint_group.id]
        verifylist = [('endpoint_group', [self._endpoint_group.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_vpn_endpoint_group.assert_called_once_with(
            self._endpoint_group.id
        )
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_e_g(*args, **kwargs):
            return _vpn_endpoint_group.VpnEndpointGroup(**{'id': args[0]})

        self.network_client.find_vpn_endpoint_group.side_effect = _mock_e_g

        arglist = ['target1', 'target2']
        verifylist = [('endpoint_group', ['target1', 'target2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(
            2, self.network_client.delete_vpn_endpoint_group.call_count
        )
        for idx, reference in enumerate(['target1', 'target2']):
            actual = ''.join(
                self.network_client.delete_vpn_endpoint_group.call_args_list[
                    idx
                ][0]
            )
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        arglist = ['target1']
        verifylist = [('endpoint_group', ['target1'])]

        self.network_client.find_vpn_ipsec_policy.side_effect = [
            'target1',
            exceptions.CommandError,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestListEndpointGroup(TestEndpointGroup):
    def setUp(self):
        super().setUp()
        self.cmd = endpoint_group.ListEndpointGroup(self.app, None)

        self.short_header = (
            'ID',
            'Name',
            'Type',
            'Endpoints',
        )

        self.short_data = (
            self._endpoint_group['id'],
            self._endpoint_group['name'],
            self._endpoint_group['type'],
            self._endpoint_group['endpoints'],
        )

        self.network_client.vpn_endpoint_groups.return_value = [
            self._endpoint_group
        ]

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_endpoint_groups.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_endpoint_groups.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetEndpointGroup(TestEndpointGroup):
    def setUp(self):
        super().setUp()

        self.network_client.update_vpn_endpoint_group.return_value = (
            self._endpoint_group
        )
        self.cmd = endpoint_group.SetEndpointGroup(self.app, None)

    def test_set_name(self):
        arglist = [self._endpoint_group.id, '--name', 'foo']
        verifylist = [
            ('endpoint_group', self._endpoint_group['id']),
            ('name', 'foo'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_endpoint_group.assert_called_once_with(
            self._endpoint_group.id, name='foo'
        )
        self.assertIsNone(result)

    def test_set_description(self):
        arglist = [self._endpoint_group.id, '--description', 'foo']
        verifylist = [
            ('endpoint_group', self._endpoint_group.id),
            ('description', 'foo'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_endpoint_group.assert_called_once_with(
            self._endpoint_group.id, description='foo'
        )
        self.assertIsNone(result)


class TestShowEndpointGroup(TestEndpointGroup):
    def setUp(self):
        super().setUp()

        self.network_client.find_vpn_endpoint_group.return_value = (
            self._endpoint_group
        )
        self.network_client.find_vpn_endpoint_group.side_effect = None
        self.cmd = endpoint_group.ShowEndpointGroup(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        arglist = [self._endpoint_group.id]
        verifylist = [('endpoint_group', self._endpoint_group.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_vpn_endpoint_group.assert_called_once_with(
            self._endpoint_group.id, ignore_missing=False
        )
        self.assertEqual(self.ordered_headers, headers)
        self.assertCountEqual(self.ordered_data, data)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Limits Action Implementation"""

import argparse
import itertools
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


def _format_absolute_limit(absolute_limits: Any) -> dict[str, Any]:
    info: dict[str, Any] = {}

    for key in set(absolute_limits):
        if key in ('id', 'name', 'location'):
            continue

        info[key] = absolute_limits[key]

    return info


def _format_rate_limit(rate_limits: Any) -> Any:
    # flatten this:
    #
    #   {'uri': '<uri>', 'limit': [{'value': '<value>', ...], ...}
    #
    # to this:
    #
    #   {'uri': '<uri>', 'value': '<value>', ...}, ...}
    return itertools.chain(
        *[[{'uri': x['uri'], **y} for y in x['limit']] for x in rate_limits]
    )


class ShowLimits(command.Lister):
    _description = _("Show compute and block storage limits")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        type_group = parser.add_mutually_exclusive_group(required=True)
        type_group.add_argument(
            "--absolute",
            dest="is_absolute",
            action="store_true",
            default=False,
            help=_("Show absolute limits"),
        )
        type_group.add_argument(
            "--rate",
            dest="is_rate",
            action="store_true",
            default=False,
            help=_(
                'Show rate limits. This is not supported by the compute '
                'service since the 12.0.0 (Liberty) release and is only '
                'supported by the block storage service when the '
                'rate-limiting middleware is enabled. It is therefore a no-op '
                'in most deployments.'
            ),
        )
        parser.add_argument(
            "--reserved",
            dest="is_reserved",
            action="store_true",
            default=False,
            help=_("Include reservations count (only valid with --absolute)"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Show limits for a specific project (name or ID) '
                '(only valid with --absolute)'
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        # deprecated alias for --project-domain
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            dest='project_domain',
            help=argparse.SUPPRESS,
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        project_id = None
        if parsed_args.project is not None:
            identity_client = sdk_utils.ensure_service_version(
                self.app.client_manager.sdk_connection.identity, '3'
            )
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )

        compute_limits = None
        volume_limits = None

        if self.app.client_manager.is_compute_endpoint_enabled():
            compute_client = self.app.client_manager.compute
            compute_limits = compute_client.get_limits(
                reserved=parsed_args.is_reserved, tenant_id=project_id
            )

        if self.app.client_manager.is_volume_endpoint_enabled():
            volume_client = self.app.client_manager.volume
            volume_limits = volume_client.get_limits(
                project=project_id,
            )

        if parsed_args.is_absolute:
            columns = ["Name", "Value"]
            info = {}
            if compute_limits:
                info.update(_format_absolute_limit(compute_limits.absolute))
            if volume_limits:
                info.update(_format_absolute_limit(volume_limits.absolute))

            return (columns, sorted(info.items(), key=lambda x: x[0]))
        else:  # parsed_args.is_rate
            data = []
            if compute_limits:
                data.extend(_format_rate_limit(compute_limits.rate))
            if volume_limits:
                data.extend(_format_rate_limit(volume_limits.rate))
            columns = [
                "Verb",
                "URI",
                "Value",
                "Remain",
                "Unit",
                "Next Available",
            ]

            return (
                columns,
                [
                    (
                        s['verb'],
                        s['uri'],
                        s['value'],
                        s['remaining'],
                        s['unit'],
                        s.get('next-available') or s['next_available'],
                    )
                    for s in data
                ],
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstack import exceptions as sdk_exc
from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import group as _group
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import role as _role
from openstack.identity.v3 import role_assignment as _role_assignment
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes
from openstackclient.identity.v3 import role_assignment
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRoleAssignmentList(identity_fakes.TestIdentity):
    columns = (
        'Role',
        'User',
        'Group',
        'Project',
        'Domain',
        'System',
        'Inherited',
    )

    def setUp(self):
        super().setUp()

        self.user = sdk_fakes.generate_fake_resource(_user.User)
        self.group = sdk_fakes.generate_fake_resource(_group.Group)
        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.role = sdk_fakes.generate_fake_resource(_role.Role)
        self.assignment_with_project_id_and_user_id = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={'project': {'id': self.project.id}},
                user={'id': self.user.id},
            )
        )
        self.assignment_with_project_id_and_group_id = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={'project': {'id': self.project.id}},
                group={'id': self.group.id},
            )
        )
        self.assignment_with_domain_id_and_user_id = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={'domain': {'id': self.domain.id}},
                user={'id': self.user.id},
            )
        )
        self.assignment_with_domain_id_and_group_id = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={'domain': {'id': self.domain.id}},
                group={'id': self.group.id},
            )
        )

        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.find_group.return_value = self.group
        self.identity_sdk_client.find_project.return_value = self.project
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_role.return_value = self.role

        # Get the command object to test
        self.cmd = role_assignment.ListRoleAssignment(self.app, None)

    def test_role_assignment_list_no_filters(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_project_id_and_user_id,
            self.assignment_with_project_id_and_group_id,
            self.assignment_with_domain_id_and_user_id,
            self.assignment_with_domain_id_and_group_id,
        ]

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                self.project.id,
                '',
                '',
                False,
            ),
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                '',
                self.domain.id,
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_user(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_user_id,
            self.assignment_with_project_id_and_user_id,
        ]

        arglist = ['--user', self.user.name]
        verifylist = [
            ('user', self.user.name),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_user.assert_called_with(
            name_or_id=self.user.name, ignore_missing=False, domain_id=None
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=self.user.id,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_user_with_domain(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_user_id,
            self.assignment_with_project_id_and_user_id,
        ]

        arglist = ['--user', self.user.name, '--user-domain', self.domain.name]
        verifylist = [
            ('user', self.user.name),
            ('user_domain', self.domain.name),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_domain.assert_called_with(
            name_or_id=self.domain.name, ignore_missing=False
        )
        self.identity_sdk_client.find_user.assert_called_with(
            name_or_id=self.user.name,
            ignore_missing=False,
            domain_id=self.domain.id,
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=self.user.id,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    @mock.patch.object(_user.User, 'find')
    def test_role_assignment_list_user_not_found(self, find_mock):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_user_id,
            self.assignment_with_project_id_and_user_id,
        ]
        self.identity_sdk_client.find_user.side_effect = [
            sdk_exc.ForbiddenException,
            sdk_exc.ForbiddenException,
        ]

        arglist = ['--user', self.user.id]
        verifylist = [
            ('user', self.user.id),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=self.user.id,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_group(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_group_id,
            self.assignment_with_project_id_and_group_id,
        ]

        arglist = ['--group', self.group.name]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', self.group.name),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_group.assert_called_with(
            name_or_id=self.group.name, ignore_missing=False, domain_id=None
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=self.group.id,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                '',
                self.group.id,
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_group_with_domain(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_group_id,
            self.assignment_with_project_id_and_group_id,
        ]

        arglist = [
            '--group',
            self.group.name,
            '--group-domain',
            self.domain.name,
        ]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', self.group.name),
            ('group_domain', self.domain.name),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_domain.assert_called_with(
            name_or_id=self.domain.name, ignore_missing=False
        )
        self.identity_sdk_client.find_group.assert_called_with(
            name_or_id=self.group.name,
            ignore_missing=False,
            domain_id=self.domain.id,
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=self.group.id,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                '',
                self.group.id,
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_domain(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_domain_id_and_user_id,
            self.assignment_with_domain_id_and_group_id,
        ]

        arglist = ['--domain', self.domain.name]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', self.domain.name),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=self.domain.id,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                '',
                self.domain.id,
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_project(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_project_id_and_user_id,
            self.assignment_with_project_id_and_group_id,
        ]

        arglist = ['--project', self.project.name]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', self.project.name),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            name_or_id=self.project.name, ignore_missing=False, domain_id=None
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=self.project.id,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_project_with_domain(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_project_id_and_user_id,
            self.assignment_with_project_id_and_group_id,
        ]

        arglist = [
            '--project',
            self.project.name,
            '--project-domain',
            self.domain.name,
        ]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', self.project.name),
            ('project_domain', self.domain.name),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_domain.assert_called_with(
            name_or_id=self.domain.name, ignore_missing=False
        )
        self.identity_sdk_client.find_project.assert_called_with(
            name_or_id=self.project.name,
            ignore_missing=False,
            domain_id=self.domain.id,
        )
        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=self.project.id,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_def_creds(self):
        self.app.client_manager.auth_ref = mock.Mock()
        auth_ref = self.app.client_manager.auth_ref
        auth_ref.project_id.return_value = self.project.id
        auth_ref.user_id.return_value = self.user.id

        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_project_id_and_user_id,
        ]

        arglist = [
            '--auth-user',
            '--auth-project',
        ]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', False),
            ('authuser', True),
            ('authproject', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=self.user.id,
            group_id=None,
            scope_project_id=self.project.id,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_effective(self):
        self.identity_sdk_client.role_assignments.return_value = [
            self.assignment_with_project_id_and_user_id,
            self.assignment_with_domain_id_and_user_id,
        ]

        arglist = ['--effective']
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', True),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=True,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
            (
                self.role.id,
                self.user.id,
                '',
                '',
                self.domain.id,
                '',
                False,
            ),
        )
        self.assertEqual(tuple(data), datalist)

    def test_role_assignment_list_inherited(self):
        assignment_with_project_id_and_user_id_inherited = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={
                    'project': {'id': self.project.id},
                    'OS-INHERIT:inherited_to': 'projects',
                },
                user={'id': self.user.id},
            )
        )
        assignment_with_domain_id_and_group_id_inherited = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id},
                scope={
                    'domain': {'id': self.domain.id},
                    'OS-INHERIT:inherited_to': 'projects',
                },
                group={'id': self.group.id},
            )
        )
        self.identity_sdk_client.role_assignments.return_value = [
            assignment_with_project_id_and_user_id_inherited,
            assignment_with_domain_id_and_group_id_inherited,
        ]

        arglist = ['--inherited']
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', True),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to='projects',
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.role.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                True,
            ),
            (
                self.role.id,
                '',
                self.group.id,
                '',
                self.domain.id,
                '',
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_include_names(self):
        assignment_with_project_id_and_user_id_include_names = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id, 'name': self.role.name},
                scope={
                    'project': {
                        'domain': {
                            'id': self.domain.id,
                            'name': self.domain.name,
                        },
                        'id': self.project.id,
                        'name': self.project.name,
                    }
                },
                user={
                    'domain': {'id': self.domain.id, 'name': self.domain.name},
                    'id': self.user.id,
                    'name': self.user.name,
                },
            )
        )
        assignment_with_domain_id_and_group_id_include_names = (
            sdk_fakes.generate_fake_resource(
                resource_type=_role_assignment.RoleAssignment,
                role={'id': self.role.id, 'name': self.role.name},
                scope={
                    'domain': {'id': self.domain.id, 'name': self.domain.name}
                },
                group={
                    'domain': {'id': self.domain.id, 'name': self.domain.name},
                    'id': self.group.id,
                    'name': self.group.name,
                },
            )
        )

        self.identity_sdk_client.role_assignments.return_value = [
            assignment_with_project_id_and_user_id_include_names,
            assignment_with_domain_id_and_group_id_include_names,
        ]

        arglist = ['--names']
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', None),
            ('effective', None),
            ('inherited', False),
            ('names', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples

        # This test will not run correctly until the patch in the python
        # client is merged. Once that is done 'data' should return the
        # correct information
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=None,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=True,
            inherited_to=None,
        )

        collist = (
            'Role',
            'User',
            'Group',
            'Project',
            'Domain',
            'System',
            'Inherited',
        )
        self.assertEqual(columns, collist)

        datalist = (
            (
                self.role.name,
                '@'.join([self.user.name, self.domain.name]),
                '',
                '@'.join([self.project.name, self.domain.name]),
                '',
                '',
                False,
            ),
            (
                self.role.name,
                '',
                '@'.join([self.group.name, self.domain.name]),
                '',
                self.domain.name,
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_domain_role(self):
        domain_2 = sdk_fakes.generate_fake_resource(_domain.Domain)
        # Create new role with same name but different domain
        role_2 = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=domain_2.id,
            name=self.role.name,
        )
        assignment_with_role_domain_2 = sdk_fakes.generate_fake_resource(
            resource_type=_role_assignment.RoleAssignment,
            role={'id': role_2.id, 'name': role_2.name},
            scope={'project': {'id': self.project.id}},
            user={'id': self.user.id},
        )

        self.identity_sdk_client.find_domain.return_value = domain_2
        self.identity_sdk_client.find_role.return_value = role_2
        self.identity_sdk_client.role_assignments.return_value = [
            assignment_with_role_domain_2,
        ]

        arglist = [
            '--role',
            role_2.name,
            '--role-domain',
            domain_2.name,
        ]
        verifylist = [
            ('user', None),
            ('user_domain', None),
            ('group', None),
            ('group_domain', None),
            ('system', None),
            ('domain', None),
            ('project', None),
            ('project_domain', None),
            ('role', role_2.name),
            ('effective', None),
            ('inherited', False),
            ('names', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.role_assignments.assert_called_with(
            role_id=role_2.id,
            user_id=None,
            group_id=None,
            scope_project_id=None,
            scope_domain_id=None,
            scope_system=None,
            effective=None,
            include_names=None,
            inherited_to=None,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                role_2.id,
                self.user.id,
                '',
                self.project.id,
                '',
                '',
                False,
            ),
        )
        self.assertEqual(datalist, tuple(data))



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from keystoneauth1 import discover
from openstack.block_storage import v2 as block_storage_v2
from openstack.image import v2 as image_v2

from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.volume = mock.Mock(spec=block_storage_v2.Proxy)
        self.app.client_manager.volume.api_version = '2'
        self.volume_client = self.app.client_manager.volume
        self.set_volume_api_version()  # default to the lowest

    def set_volume_api_version(self, version: str | None = None):
        """Set a fake block storage API version.

        :param version: The fake microversion to "support". This must be None
            since cinder v2 didn't support microversions.
        :returns: None
        """
        assert version is None

        self.volume_client.default_microversion = None
        self.volume_client.get_endpoint_data.return_value = (
            discover.EndpointData(
                min_microversion=None,
                max_microversion=None,
            )
        )


class TestVolume(
    identity_fakes.FakeClientMixin,
    FakeClientMixin,
    utils.TestCommand,
):
    def setUp(self):
        super().setUp()

        # avoid circular imports by defining this manually rather than using
        # openstackclient.tests.unit.image.v2.fakes.FakeClientMixin
        # TODO(stephenfin): Switch to spec_set once keystoneauth exposes
        # instance attributes as class attributes
        # https://review.opendev.org/c/openstack/keystoneauth/+/994090
        self.app.client_manager.image = mock.Mock(spec=image_v2.Proxy)
        self.image_client = self.app.client_manager.image



# Copyright (c) 2020 Troila.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Router NDP proxy action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import ndp_proxy as _ndp_proxy
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


def _get_columns(
    item: _ndp_proxy.NDPProxy,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateNDPProxy(command.ShowOne):
    _description = _("Create NDP proxy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'router', metavar='<router>', help=_("The name or ID of a router")
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_("New NDP proxy name")
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            required=True,
            help=_(
                "The name or ID of the network port associated "
                "to the NDP proxy"
            ),
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help=_(
                "The IPv6 address that is to be proxied. In case the port "
                "has multiple addresses assigned, use this option to "
                "select which address is to be used."
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_(
                "Text to describe/contextualize the use of the "
                "NDP proxy configuration"
            ),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        attrs = {'name': parsed_args.name}
        client = self.app.client_manager.network
        router = client.find_router(
            parsed_args.router,
            ignore_missing=False,
        )
        attrs['router_id'] = router.id

        if parsed_args.ip_address:
            attrs['ip_address'] = parsed_args.ip_address

        port = client.find_port(parsed_args.port, ignore_missing=False)
        attrs['port_id'] = port.id

        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description

        obj = client.create_ndp_proxy(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)


class DeleteNDPProxy(command.Command):
    _description = _("Delete NDP proxy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ndp_proxy',
            nargs="+",
            metavar="<ndp-proxy>",
            help=_("NDP proxy(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for ndp_proxy in parsed_args.ndp_proxy:
            try:
                obj = client.find_ndp_proxy(ndp_proxy, ignore_missing=False)
                client.delete_ndp_proxy(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to delete NDP proxy '%(ndp_proxy)s': %(e)s"),
                    {'ndp_proxy': ndp_proxy, 'e': e},
                )
        if result > 0:
            total = len(parsed_args.ndp_proxy)
            msg = _(
                "%(result)s of %(total)s NDP proxies failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListNDPProxy(command.Lister):
    _description = _("List NDP proxies")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--router',
            metavar='<router>',
            help=_(
                "List only NDP proxies associated with the specifed router "
                "(name or ID)"
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_(
                "List only NDP proxies associated with the specified port "
                "(name or ID)"
            ),
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help=_(
                "List only NDP proxies associated with the specified "
                "IPv6 address"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only NDP proxies with the specified project (name or ID)"
            ),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_("List only NDP proxies with the specified name"),
        )

        identity_common.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = (
            'id',
            'name',
            'router_id',
            'ip_address',
            'project_id',
        )
        headers = (
            'ID',
            'Name',
            'Router ID',
            'IP Address',
            'Project',
        )

        query = {}

        if parsed_args.router:
            router = client.find_router(
                parsed_args.router, ignore_missing=False
            )
            query['router_id'] = router.id
        if parsed_args.port:
            port = client.find_port(parsed_args.port, ignore_missing=False)
            query['port_id'] = port.id
        if parsed_args.ip_address is not None:
            query['ip_address'] = parsed_args.ip_address
        if parsed_args.project:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            query['project_id'] = project_id
        if parsed_args.name:
            query['name'] = parsed_args.name

        data = client.ndp_proxies(**query)

        return (
            headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetNDPProxy(command.Command):
    _description = _("Set NDP proxy properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ndp_proxy',
            metavar='<ndp-proxy>',
            help=_("The ID or name of the NDP proxy to update"),
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_("Set NDP proxy name")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_(
                "Text to describe/contextualize the use of "
                "the NDP proxy configuration"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        attrs = {}
        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description
        if parsed_args.name is not None:
            attrs['name'] = parsed_args.name

        obj = client.find_ndp_proxy(
            parsed_args.ndp_proxy, ignore_missing=False
        )
        client.update_ndp_proxy(obj, **attrs)


class ShowNDPProxy(command.ShowOne):
    _description = _("Display NDP proxy details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ndp_proxy',
            metavar="<ndp-proxy>",
            help=_("The ID or name of the NDP proxy"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_ndp_proxy(
            parsed_args.ndp_proxy, ignore_missing=False
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
import io
from unittest import mock

from keystoneauth1 import session
from osc_lib import exceptions
from requests_mock.contrib import fixture

from openstackclient.api import object_store_v1 as object_store
from openstackclient.object.v1 import object as object_cmds
from openstackclient.tests.unit.object.v1 import fakes as object_fakes


class TestObjectAll(object_fakes.TestObjectV1):
    def setUp(self):
        super().setUp()

        # these tests require a "real" client since we mock requests
        self.requests_mock = self.useFixture(fixture.Fixture())
        self.app.client_manager.object_store = object_store.APIv1(
            session=session.Session(),
            endpoint=object_fakes.ENDPOINT,
        )


class TestObjectCreate(TestObjectAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = object_cmds.CreateObject(self.app, None)

    def test_multiple_object_create_with_object_name(self):
        arglist = [
            object_fakes.container_name,
            object_fakes.object_name_1,
            object_fakes.object_name_2,
            '--name',
            object_fakes.object_upload_name,
        ]

        verifylist = [
            ('container', object_fakes.container_name),
            (
                'objects',
                [object_fakes.object_name_1, object_fakes.object_name_2],
            ),
            ('name', object_fakes.object_upload_name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestObjectList(TestObjectAll):
    columns = ('Name',)

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = object_cmds.ListObject(self.app, None)

    def test_object_list_objects_no_options(self):
        return_body = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]
        self.requests_mock.register_uri(
            'GET',
            object_fakes.ENDPOINT
            + '/'
            + object_fakes.container_name
            + '?format=json',
            json=return_body,
            status_code=200,
        )

        arglist = [
            object_fakes.container_name,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Lister.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (object_fakes.object_name_1,),
            (object_fakes.object_name_2,),
        ]
        self.assertEqual(datalist, list(data))

    def test_object_list_objects_prefix(self):
        return_body = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]
        self.requests_mock.register_uri(
            'GET',
            object_fakes.ENDPOINT
            + '/'
            + object_fakes.container_name_2
            + '?prefix=floppy&format=json',
            json=return_body,
            status_code=200,
        )

        arglist = [
            '--prefix',
            'floppy',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('prefix', 'floppy'),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = ((object_fakes.object_name_2,),)
        self.assertEqual(datalist, tuple(data))


class TestObjectShow(TestObjectAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = object_cmds.ShowObject(self.app, None)

    def test_object_show(self):
        headers = {
            'content-type': 'text/plain',
            'content-length': '20',
            'last-modified': 'yesterday',
            'etag': '4c4e39a763d58392724bccf76a58783a',
            'x-container-meta-owner': object_fakes.ACCOUNT_ID,
            'x-object-manifest': 'manifest',
        }
        self.requests_mock.register_uri(
            'HEAD',
            '/'.join(
                [
                    object_fakes.ENDPOINT,
                    object_fakes.container_name,
                    object_fakes.object_name_1,
                ]
            ),
            headers=headers,
            status_code=200,
        )

        arglist = [
            object_fakes.container_name,
            object_fakes.object_name_1,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
            ('object', object_fakes.object_name_1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            'account',
            'container',
            'content-length',
            'content-type',
            'etag',
            'last-modified',
            'object',
            'x-object-manifest',
        )
        self.assertEqual(collist, columns)
        datalist = (
            object_fakes.ACCOUNT_ID,
            object_fakes.container_name,
            '20',
            'text/plain',
            '4c4e39a763d58392724bccf76a58783a',
            'yesterday',
            object_fakes.object_name_1,
            'manifest',
        )
        self.assertEqual(datalist, data)


class TestObjectSave(TestObjectAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = object_cmds.SaveObject(self.app, None)

    def test_save_to_stdout(self):
        self.requests_mock.register_uri(
            'GET',
            object_fakes.ENDPOINT
            + '/'
            + object_fakes.container_name
            + '/'
            + object_fakes.object_name_1,
            status_code=200,
            content=object_fakes.object_1_content,
        )

        arglist = [
            object_fakes.container_name,
            object_fakes.object_name_1,
            '--file',
            '-',
        ]

        verifylist = [
            ('container', object_fakes.container_name),
            ('object', object_fakes.object_name_1),
            ('file', '-'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        class FakeStdout(io.BytesIO):
            def __init__(self):
                io.BytesIO.__init__(self)
                self.context_manager_calls = []

            def __enter__(self):
                self.context_manager_calls.append('__enter__')
                return self

            def __exit__(self, *a):
                self.context_manager_calls.append('__exit__')

        with (
            mock.patch('sys.stdout') as fake_stdout,
            mock.patch('os.fdopen', return_value=FakeStdout()) as fake_fdopen,
        ):
            fake_stdout.fileno.return_value = 123
            self.cmd.take_action(parsed_args)

        self.assertEqual(
            fake_fdopen.return_value.getvalue(), object_fakes.object_1_content
        )
        self.assertEqual(fake_fdopen.mock_calls, [mock.call(123, 'wb')])
        self.assertEqual(
            fake_fdopen.return_value.context_manager_calls,
            ['__enter__', '__exit__'],
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Identity v3 Service Catalog action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from cliff import columns as cliff_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class EndpointsColumn(cliff_columns.FormattableColumn[Any]):
    def human_readable(self) -> str:
        if not self._value:
            return ""
        ret = ''
        for ep in self._value:
            region = ep.get('region_id') or ep.get('region') or '<none>'
            ret += region + '\n'
            ret += "  {}: {}\n".format(ep['interface'], ep['url'])
        return ret


class ListCatalog(command.Lister):
    _description = _("List services in the service catalog")

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        # Trigger auth if it has not happened yet
        auth_ref = self.app.client_manager.auth_ref
        if not auth_ref:
            raise exceptions.AuthorizationFailure(
                "Only an authorized user may issue a new token."
            )

        data = auth_ref.service_catalog.catalog
        columns = ('Name', 'Type', 'Endpoints')
        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters={
                        'Endpoints': EndpointsColumn,
                    },
                )
                for s in data
            ),
        )


class ShowCatalog(command.ShowOne):
    _description = _("Display service catalog details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Service to display (type or name)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        # Trigger auth if it has not happened yet
        auth_ref = self.app.client_manager.auth_ref
        if not auth_ref:
            raise exceptions.AuthorizationFailure(
                "Only an authorized user may issue a new token."
            )

        data = None
        for service in auth_ref.service_catalog.catalog:
            if (
                service.get('name') == parsed_args.service
                or service.get('type') == parsed_args.service
            ):
                data = dict(service)
                data['endpoints'] = EndpointsColumn(data['endpoints'])
                if 'links' in data:
                    data.pop('links')
                break

        if not data:
            LOG.error(_('service %s not found\n'), parsed_args.service)
            return ((), ())

        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Block Storage Service action implementations"""

import argparse
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions

from openstackclient import command
from openstackclient.i18n import _


class BlockStorageLogLevelList(command.Lister):
    """List log levels of block storage service.

    Supported by --os-volume-api-version 3.32 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--host",
            metavar="<host>",
            default=None,
            help=_(
                "List block storage service log level of specified host "
                "(name only)"
            ),
        )
        parser.add_argument(
            "--service",
            metavar="<service>",
            default=None,
            choices=(
                None,
                '*',
                'cinder-api',
                'cinder-volume',
                'cinder-scheduler',
                'cinder-backup',
            ),
            help=_(
                "List block storage service log level of the specified "
                "service (name only)"
            ),
        )
        parser.add_argument(
            "--log-prefix",
            metavar="<log-prefix>",
            default=None,
            help="Prefix for the log, e.g. 'sqlalchemy'",
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[list[str], list[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        columns = [
            "Binary",
            "Host",
            "Prefix",
            "Level",
        ]

        if not sdk_utils.supports_microversion(volume_client, '3.32'):
            msg = _(
                "--os-volume-api-version 3.32 or greater is required to "
                "support the 'block storage log level list' command"
            )
            raise exceptions.CommandError(msg)

        data = []
        for entry in volume_client.get_service_log_levels(
            binary=parsed_args.service,
            server=parsed_args.host,
            prefix=parsed_args.log_prefix,
        ):
            entry_levels = sorted(entry.levels.items(), key=lambda x: x[0])
            for prefix, level in entry_levels:
                data.append((entry.binary, entry.host, prefix, level))

        return (columns, data)


class BlockStorageLogLevelSet(command.Command):
    """Set log level of block storage service

    Supported by --os-volume-api-version 3.32 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "level",
            metavar="<log-level>",
            choices=('INFO', 'WARNING', 'ERROR', 'DEBUG'),
            type=str.upper,
            help=_("Desired log level"),
        )
        parser.add_argument(
            "--host",
            metavar="<host>",
            default=None,
            help=_(
                "Set block storage service log level of specified host "
                "(name only)"
            ),
        )
        parser.add_argument(
            "--service",
            metavar="<service>",
            default=None,
            choices=(
                None,
                '*',
                'cinder-api',
                'cinder-volume',
                'cinder-scheduler',
                'cinder-backup',
            ),
            help=_(
                "Set block storage service log level of specified service "
                "(name only)"
            ),
        )
        parser.add_argument(
            "--log-prefix",
            metavar="<log-prefix>",
            default=None,
            help="Prefix for the log, e.g. 'sqlalchemy'",
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.32'):
            msg = _(
                "--os-volume-api-version 3.32 or greater is required to "
                "support the 'block storage log level set' command"
            )
            raise exceptions.CommandError(msg)

        volume_client.set_service_log_levels(
            level=parsed_args.level,
            binary=parsed_args.service,
            server=parsed_args.host,
            prefix=parsed_args.log_prefix,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Volume V3 Resource Filters implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class ListBlockStorageResourceFilter(command.Lister):
    _description = _('List block storage resource filters')

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.33'):
            msg = _(
                "--os-volume-api-version 3.33 or greater is required to "
                "support the 'block storage resource filter list' command"
            )
            raise exceptions.CommandError(msg)

        column_headers = (
            'Resource',
            'Filters',
        )
        columns = (
            'resource',
            'filters',
        )

        data = volume_client.resource_filters()
        formatters = {'filters': format_columns.ListColumn}

        return (
            column_headers,
            (
                utils.get_item_properties(s, columns, formatters=formatters)
                for s in data
            ),
        )


class ShowBlockStorageResourceFilter(command.ShowOne):
    _description = _('Show filters for a block storage resource type')

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'resource',
            metavar='<resource>',
            help=_('Resource to show filters for (name).'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.33'):
            msg = _(
                "--os-volume-api-version 3.33 or greater is required to "
                "support the 'block storage resource filter show' command"
            )
            raise exceptions.CommandError(msg)

        data = volume_client.resource_filters(resource=parsed_args.resource)
        if not data:
            msg = _(
                "No resource filter with a name of {parsed_args.resource}' "
                "exists."
            )
            raise exceptions.CommandError(msg)
        resource_filter = next(data)

        column_headers = (
            'Resource',
            'Filters',
        )
        columns = (
            'resource',
            'filters',
        )
        formatters = {'filters': format_columns.ListColumn}

        return (
            column_headers,
            utils.get_dict_properties(
                resource_filter, columns, formatters=formatters
            ),
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.volume.v2 import common


class TransferRequestTests(common.BaseVolumeTests):
    """Functional tests for transfer request."""

    API_VERSION = '2'

    def test_volume_transfer_request_accept(self):
        volume_name = uuid.uuid4().hex
        xfer_name = uuid.uuid4().hex

        # create a volume
        cmd_output = self.openstack(
            'volume create ' + '--size 1 ' + volume_name,
            parse_output=True,
        )
        self.assertEqual(volume_name, cmd_output['name'])
        self.addCleanup(
            self.openstack,
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume delete '
            + volume_name,
        )
        self.wait_for_status("volume", volume_name, "available")

        # create volume transfer request for the volume
        # and get the auth_key of the new transfer request
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request create '
            + ' --name '
            + xfer_name
            + ' '
            + volume_name,
            parse_output=True,
        )
        self.assertEqual(xfer_name, cmd_output['name'])
        xfer_id = cmd_output['id']
        auth_key = cmd_output['auth_key']
        self.assertTrue(auth_key)
        self.wait_for_status("volume", volume_name, "awaiting-transfer")

        # accept the volume transfer request
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request accept '
            + '--auth-key '
            + auth_key
            + ' '
            + xfer_id,
            parse_output=True,
        )
        self.assertEqual(xfer_name, cmd_output['name'])
        self.wait_for_status("volume", volume_name, "available")

    def test_volume_transfer_request_list_show(self):
        volume_name = uuid.uuid4().hex
        xfer_name = uuid.uuid4().hex

        # create a volume
        cmd_output = self.openstack(
            'volume create ' + '--size 1 ' + volume_name,
            parse_output=True,
        )
        self.assertEqual(volume_name, cmd_output['name'])
        self.addCleanup(
            self.openstack,
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume delete '
            + volume_name,
        )
        self.wait_for_status("volume", volume_name, "available")

        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request create '
            + ' --name '
            + xfer_name
            + ' '
            + volume_name,
            parse_output=True,
        )
        self.assertEqual(xfer_name, cmd_output['name'])
        xfer_id = cmd_output['id']
        auth_key = cmd_output['auth_key']
        self.assertTrue(auth_key)
        self.wait_for_status("volume", volume_name, "awaiting-transfer")

        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request list',
            parse_output=True,
        )
        self.assertIn(xfer_name, [req['Name'] for req in cmd_output])

        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request show '
            + xfer_id,
            parse_output=True,
        )
        self.assertEqual(xfer_name, cmd_output['name'])

        # NOTE(dtroyer): We need to delete the transfer request to allow the
        #                volume to be deleted. The addCleanup() route does
        #                not have a mechanism to wait for the volume status
        #                to become 'available' before attempting to  delete
        #                the volume.
        cmd_output = self.openstack(
            '--os-volume-api-version '
            + self.API_VERSION
            + ' '
            + 'volume transfer request delete '
            + xfer_id
        )
        self.wait_for_status("volume", volume_name, "available")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock

from openstackclient.network.v2.dynamic_routing import bgp_peer
from openstackclient.tests.unit.network.v2.dynamic_routing import fakes
from openstackclient.tests.unit.network.v2 import fakes as test_fakes


class TestListBgpPeer(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self._bgp_peers = fakes.FakeBgpPeer.create_bgp_peers(count=1)
        self.columns = ('ID', 'Name', 'Peer IP', 'Remote AS')
        self.data = []
        for _bgp_peer in self._bgp_peers:
            self.data.append(
                (
                    _bgp_peer['id'],
                    _bgp_peer['name'],
                    _bgp_peer['peer_ip'],
                    _bgp_peer['remote_as'],
                )
            )

        self.network_client.bgp_peers.return_value = self._bgp_peers

        # Get the command object to test
        self.cmd = bgp_peer.ListBgpPeer(self.app, None)

    def test_bgp_peer_list(self):
        parsed_args = self.check_parser(self.cmd, [], [])

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.bgp_peers.assert_called_once_with(
            retrieve_all=True
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestDeleteBgpPeer(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self._bgp_peer = fakes.FakeBgpPeer.create_one_bgp_peer()

        self.network_client.delete_bgp_peer.return_value = None
        self.network_client.find_bgp_peer.return_value = mock.Mock(
            id=self._bgp_peer.id
        )
        self.cmd = bgp_peer.DeleteBgpPeer(self.app, None)

    def test_delete_bgp_peer(self):
        arglist = [
            self._bgp_peer['name'],
        ]
        verifylist = [
            ('bgp_peer', self._bgp_peer['name']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_bgp_peer.assert_called_once_with(
            self._bgp_peer.id
        )
        self.assertIsNone(result)


class TestShowBgpPeer(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self._one_bgp_peer = fakes.FakeBgpPeer.create_one_bgp_peer()
        self.data = (
            self._one_bgp_peer['auth_type'],
            self._one_bgp_peer['id'],
            self._one_bgp_peer['name'],
            self._one_bgp_peer['peer_ip'],
            self._one_bgp_peer['tenant_id'],
            self._one_bgp_peer['remote_as'],
        )
        self._bgp_peer = self._one_bgp_peer
        self._bgp_peer_name = self._one_bgp_peer['name']
        self.columns = (
            'auth_type',
            'id',
            'name',
            'peer_ip',
            'project_id',
            'remote_as',
        )

        self.network_client.find_bgp_peer.return_value = self._bgp_peer
        # Get the command object to test
        self.cmd = bgp_peer.ShowBgpPeer(self.app, None)

    def test_bgp_peer_show(self):
        arglist = [
            self._bgp_peer_name,
        ]
        verifylist = [
            ('bgp_peer', self._bgp_peer_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        data = self.cmd.take_action(parsed_args)
        self.network_client.find_bgp_peer.assert_called_once_with(
            self._bgp_peer.name, ignore_missing=False
        )
        self.assertEqual(self.columns, data[0])
        self.assertEqual(self.data, data[1])


class TestSetBgpPeer(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self._one_bgp_peer = fakes.FakeBgpPeer.create_one_bgp_peer()
        self._bgp_peer_name = self._one_bgp_peer['name']

        self.network_client.update_bgp_peer.return_value = None
        self.network_client.find_bgp_peer.return_value = mock.Mock(
            id=self._one_bgp_peer.id
        )

        self.cmd = bgp_peer.SetBgpPeer(self.app, None)

    def test_set_bgp_peer(self):
        arglist = [
            self._bgp_peer_name,
            '--name',
            'noob',
        ]
        verifylist = [
            ('bgp_peer', self._bgp_peer_name),
            ('name', 'noob'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {'name': 'noob', 'password': None}
        self.network_client.update_bgp_peer.assert_called_once_with(
            self._one_bgp_peer.id, **attrs
        )
        self.assertIsNone(result)



#!/usr/bin/env python
# osc-lib.py - Example using OSC as a library

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
OpenStackClient Library Examples

This script shows the basic use of the OpenStackClient ClientManager
as a library.

"""

import argparse
import logging
import sys

import common
from os_client_config import config as cloud_config

from openstackclient.common import clientmanager


LOG = logging.getLogger('')


def run(opts):
    """Run the examples"""

    # Do configuration file handling
    cc = cloud_config.OpenStackConfig()
    LOG.debug("defaults: %s", cc.defaults)

    cloud = cc.get_one_cloud(
        cloud=opts.cloud,
        argparse=opts,
    )
    LOG.debug("cloud cfg: %s", cloud.config)

    # Loop through extensions to get API versions
    # Currently API versions are statically selected.  Once discovery
    # is working this can go away...
    api_version = {}
    for mod in clientmanager.PLUGIN_MODULES:
        version_opt = getattr(opts, mod.API_VERSION_OPTION, None)
        if version_opt:
            api = mod.API_NAME
            api_version[api] = version_opt

    # Set up certificate verification and CA bundle
    # NOTE(dtroyer): This converts from the usual OpenStack way to the single
    #                requests argument and is an app-specific thing because
    #                we want to be like OpenStackClient.
    if opts.cacert:
        verify = opts.cacert
    else:
        verify = not opts.insecure

    # Get a ClientManager
    # Collect the auth and config options together and give them to
    # ClientManager and it will wrangle all of the goons into place.
    client_manager = clientmanager.ClientManager(
        cli_options=cloud,
        verify=verify,
        api_version=api_version,
    )

    # At this point we have a working client manager with a configured
    # session and authentication plugin.  From here on it is the app
    # making the decisions.  Need to talk to two clouds?  Make another
    # client manager with different opts.  Or use a config file and load it
    # directly into the plugin.  This example doesn't show that (yet).

    # Do useful things with it

    # Look in the object store
    c_list = client_manager.object_store.container_list()
    print("Name\tCount\tBytes")
    for c in c_list:
        print(f"{c['name']}\t{c['count']}\t{c['bytes']}")

    if len(c_list) > 0:
        # See what is in the first container
        o_list = client_manager.object_store.object_list(c_list[0]['name'])
        print("\nObject")
        for o in o_list:
            print(f"{o}")

    # Look at the compute flavors
    flavor_list = client_manager.compute.flavors.list()
    print("\nFlavors:")
    for f in flavor_list:
        print(f"{f}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='ClientManager Example')
    opts = common.base_parser(
        clientmanager.build_plugin_option_parser(parser),
    ).parse_args()

    common.configure_logging(opts)
    sys.exit(common.main(opts, run))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import ddt
from manilaclient import api_versions
from manilaclient.osc import utils
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient.share.v2 import services as services
from openstackclient.tests.unit.share.v2 import fakes as share_fakes


class TestShareService(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.services_mock = self.share_client.services
        self.services_mock.reset_mock()


class TestShareServiceSet(TestShareService):
    def setUp(self):
        super().setUp()

        self.share_service = share_fakes.FakeShareService.create_fake_service()
        self.cmd = services.SetShareService(self.app, None)

    def test_share_service_set_enable(self):
        arglist = [
            self.share_service.host,
            self.share_service.binary,
            '--enable',
        ]
        verifylist = [
            ('host', self.share_service.host),
            ('binary', self.share_service.binary),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.services_mock.enable.assert_called_with(
            self.share_service.host, self.share_service.binary
        )
        self.assertIsNone(result)

    def test_share_service_set_enable_exception(self):
        arglist = [
            self.share_service.host,
            self.share_service.binary,
            '--enable',
        ]
        verifylist = [
            ('host', self.share_service.host),
            ('binary', self.share_service.binary),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.services_mock.enable.side_effect = Exception()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_service_set_disable(self):
        arglist = [
            self.share_service.host,
            self.share_service.binary,
            '--disable',
        ]
        verifylist = [
            ('host', self.share_service.host),
            ('binary', self.share_service.binary),
            ('disable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.services_mock.disable.assert_called_with(
            self.share_service.host, self.share_service.binary
        )
        self.assertIsNone(result)

    def test_service_set_disable_with_reason(self):
        self.set_share_api_version('2.83')
        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason',
            reason,
            self.share_service.host,
            self.share_service.binary,
        ]
        verifylist = [
            ('host', self.share_service.host),
            ('binary', self.share_service.binary),
            ('disable', True),
            ('disable_reason', reason),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.services_mock.disable.assert_called_with(
            self.share_service.host,
            self.share_service.binary,
            disable_reason=reason,
        )
        self.assertIsNone(result)

    def test_share_service_set_disable_exception(self):
        arglist = [
            self.share_service.host,
            self.share_service.binary,
            '--disable',
        ]
        verifylist = [
            ('host', self.share_service.host),
            ('binary', self.share_service.binary),
            ('disable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.services_mock.disable.side_effect = Exception()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


@ddt.ddt
class TestShareServiceList(TestShareService):
    columns = ['id', 'binary', 'host', 'zone', 'status', 'state', 'updated_at']
    columns_with_reason = [*columns, 'disabled_reason']

    column_headers = utils.format_column_headers(columns)
    column_headers_with_reason = utils.format_column_headers(
        columns_with_reason
    )

    def setUp(self):
        super().setUp()

        self.services_list = share_fakes.FakeShareService.create_fake_services(
            {'disabled_reason': ''}
        )
        self.services_mock.list.return_value = self.services_list
        self.values = (
            oscutils.get_dict_properties(i._info, self.columns)
            for i in self.services_list
        )
        self.values_with_reason = (
            oscutils.get_dict_properties(i._info, self.columns_with_reason)
            for i in self.services_list
        )

        self.cmd = services.ListShareService(self.app, None)

    @ddt.data('2.82', '2.83')
    def test_share_service_list(self, version):
        self.set_share_api_version(version)
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.services_mock.list.assert_called_with(
            search_opts={
                'host': None,
                'binary': None,
                'status': None,
                'state': None,
                'zone': None,
            }
        )
        if api_versions.APIVersion(version) >= api_versions.APIVersion("2.83"):
            self.assertEqual(self.column_headers_with_reason, columns)
            self.assertEqual(list(self.values_with_reason), list(data))
        else:
            self.assertEqual(self.column_headers, columns)
            self.assertEqual(list(self.values), list(data))

    @ddt.data('2.82', '2.83')
    def test_share_service_list_host_status(self, version):
        self.set_share_api_version(version)
        arglist = [
            '--host',
            self.services_list[0].host,
            '--status',
            self.services_list[1].status,
        ]
        verifylist = [
            ('host', self.services_list[0].host),
            ('status', self.services_list[1].status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.services_mock.list.assert_called_with(
            search_opts={
                'host': self.services_list[0].host,
                'binary': None,
                'status': self.services_list[1].status,
                'state': None,
                'zone': None,
            }
        )
        if api_versions.APIVersion(version) >= api_versions.APIVersion("2.83"):
            self.assertEqual(self.column_headers_with_reason, columns)
            self.assertEqual(list(self.values_with_reason), list(data))
        else:
            self.assertEqual(self.column_headers, columns)
            self.assertEqual(list(self.values), list(data))

    @ddt.data('2.82', '2.83')
    def test_share_service_list_binary_state_zone(self, version):
        self.set_share_api_version(version)
        arglist = [
            '--binary',
            self.services_list[0].binary,
            '--state',
            self.services_list[1].state,
            '--zone',
            self.services_list[1].zone,
        ]
        verifylist = [
            ('binary', self.services_list[0].binary),
            ('state', self.services_list[1].state),
            ('zone', self.services_list[1].zone),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.services_mock.list.assert_called_with(
            search_opts={
                'host': None,
                'binary': self.services_list[0].binary,
                'status': None,
                'state': self.services_list[1].state,
                'zone': self.services_list[1].zone,
            }
        )
        if api_versions.APIVersion(version) >= api_versions.APIVersion("2.83"):
            self.assertEqual(self.column_headers_with_reason, columns)
            self.assertEqual(list(self.values_with_reason), list(data))
        else:
            self.assertEqual(self.column_headers, columns)
            self.assertEqual(list(self.values), list(data))


@ddt.ddt
class TestShareServiceEnsureShares(TestShareService):
    def setUp(self):
        super().setUp()

        self.cmd = services.EnsureShareService(self.app, None)

    def test_ensure_shares(self):
        self.set_share_api_version('2.86')
        fake_host = 'fake_host@fakebackend'
        arglist = [
            fake_host,
        ]
        verifylist = [
            ('host', fake_host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.services_mock.ensure_shares.assert_called_with(fake_host)

    def test_ensure_shares_invalid_version(self):
        self.set_share_api_version('2.85')
        fake_host = 'fake_host@fakebackend'
        arglist = [
            fake_host,
        ]
        verifylist = [
            ('host', fake_host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_ensure_shares_command_error(self):
        self.set_share_api_version('2.86')
        self.services_mock.ensure_shares.side_effect = Exception()
        fake_host = 'fake_host@fakebackend'
        arglist = [
            fake_host,
        ]
        verifylist = [
            ('host', fake_host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )



#   Copyright 2013 OpenStack Foundation.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Usage action implementations"""

import argparse
from collections.abc import Collection, Iterable, Sequence
import datetime
import functools
from typing import Any

from cliff import columns as cliff_columns
from openstack import utils as sdk_utils
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


# TODO(stephenfin): This exists in a couple of places and should be moved to a
# common module
class ProjectColumn(cliff_columns.FormattableColumn[str]):
    """Formattable column for project column.

    Unlike the parent FormattableColumn class, the initializer of the class
    takes project_cache as the second argument.
    ``osc_lib.utils.get_item_properties`` instantiates ``FormattableColumn``
    objects with a single parameter, the column value, so you need to pass a
    partially initialized class like ``functools.partial(ProjectColumn,
    project_cache)`` to use this.
    """

    def __init__(
        self, value: str, project_cache: dict[str, str] | None = None
    ) -> None:
        super().__init__(value)
        self.project_cache = project_cache or {}

    def human_readable(self) -> str:
        project = self._value
        if not project:
            return ''

        if project in self.project_cache:
            return self.project_cache[project]

        return project


class CountColumn(cliff_columns.FormattableColumn[Collection[Any]]):
    def human_readable(self) -> str:
        return str(len(self._value)) if self._value is not None else ""


class FloatColumn(cliff_columns.FormattableColumn[float]):
    def human_readable(self) -> str:
        return f"{self._value:.2f}"


def _formatters(project_cache: dict[str, str] | None) -> dict[str, Any]:
    return {
        'project_id': functools.partial(
            ProjectColumn, project_cache=project_cache
        ),
        'server_usages': CountColumn,
        'total_memory_mb_usage': FloatColumn,
        'total_vcpus_usage': FloatColumn,
        'total_local_gb_usage': FloatColumn,
    }


def _get_usage_marker(usage: Any) -> str | None:
    marker = None
    if hasattr(usage, 'server_usages') and usage.server_usages:
        marker = usage.server_usages[-1]['instance_id']
    return marker


def _get_usage_list_marker(usage_list: Any) -> str | None:
    marker = None
    if usage_list:
        marker = _get_usage_marker(usage_list[-1])
    return marker


def _merge_usage(usage: Any, next_usage: Any) -> None:
    usage.server_usages.extend(next_usage.server_usages)
    usage.total_hours += next_usage.total_hours
    usage.total_memory_mb_usage += next_usage.total_memory_mb_usage
    usage.total_vcpus_usage += next_usage.total_vcpus_usage
    usage.total_local_gb_usage += next_usage.total_local_gb_usage


def _merge_usage_list(usages: Any, next_usage_list: Any) -> None:
    for next_usage in next_usage_list:
        if next_usage.project_id in usages:
            _merge_usage(usages[next_usage.project_id], next_usage)
        else:
            usages[next_usage.project_id] = next_usage


class ListUsage(command.Lister):
    _description = _("List resource usage per project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--start",
            metavar="<start>",
            default=None,
            help=_(
                "Usage range start date, ex 2012-01-20 (default: 4 weeks ago)"
            ),
        )
        parser.add_argument(
            "--end",
            metavar="<end>",
            default=None,
            help=_("Usage range end date, ex 2012-01-20 (default: tomorrow)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        def _format_project(project: str) -> str:
            if not project:
                return ""
            if project in project_cache:
                return project_cache[project]
            else:
                return project

        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        columns = (
            "project_id",
            "server_usages",
            "total_memory_mb_usage",
            "total_vcpus_usage",
            "total_local_gb_usage",
        )
        column_headers = (
            "Project",
            "Servers",
            "RAM MB-Hours",
            "CPU Hours",
            "Disk GB-Hours",
        )

        date_cli_format = "%Y-%m-%d"
        now = datetime.datetime.now(datetime.UTC).replace(tzinfo=None)

        if parsed_args.start:
            start = datetime.datetime.strptime(
                parsed_args.start, date_cli_format
            )
        else:
            start = now - datetime.timedelta(weeks=4)

        if parsed_args.end:
            end = datetime.datetime.strptime(parsed_args.end, date_cli_format)
        else:
            end = now + datetime.timedelta(days=1)

        usage_list = list(
            compute_client.usages(
                start=start,
                end=end,
                detailed=True,
            )
        )

        # Cache the project list
        project_cache = {}
        try:
            for p in identity_client.projects():
                project_cache[p.id] = p.name
        except Exception:  # noqa: S110
            # Just forget it if there's any trouble
            pass

        if parsed_args.formatter == 'table' and len(usage_list) > 0:
            self.app.stdout.write(
                _("Usage from %(start)s to %(end)s: \n")
                % {
                    "start": start.strftime(date_cli_format),
                    "end": end.strftime(date_cli_format),
                }
            )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters(project_cache),
                )
                for s in usage_list
            ),
        )


class ShowUsage(command.ShowOne):
    _description = _("Show resource usage for a single project")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--project",
            metavar="<project>",
            default=None,
            help=_("Name or ID of project to show usage for"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            "--start",
            metavar="<start>",
            default=None,
            help=_(
                "Usage range start date, ex 2012-01-20 (default: 4 weeks ago)"
            ),
        )
        parser.add_argument(
            "--end",
            metavar="<end>",
            default=None,
            help=_("Usage range end date, ex 2012-01-20 (default: tomorrow)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        date_cli_format = "%Y-%m-%d"
        now = datetime.datetime.now(datetime.UTC).replace(tzinfo=None)

        if parsed_args.start:
            start = datetime.datetime.strptime(
                parsed_args.start, date_cli_format
            )
        else:
            start = now - datetime.timedelta(weeks=4)

        if parsed_args.end:
            end = datetime.datetime.strptime(parsed_args.end, date_cli_format)
        else:
            end = now + datetime.timedelta(days=1)

        if parsed_args.project:
            project = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
        else:
            # Get the project from the current auth
            _project_id = self.app.client_manager.auth_ref.project_id
            assert _project_id is not None  # narrow type
            project = _project_id

        usage = compute_client.get_usage(
            project=project,
            start=start,
            end=end,
        )

        if parsed_args.formatter == 'table':
            self.app.stdout.write(
                _("Usage from %(start)s to %(end)s on project %(project)s: \n")
                % {
                    "start": start.strftime(date_cli_format),
                    "end": end.strftime(date_cli_format),
                    "project": project,
                }
            )

        columns = (
            "project_id",
            "server_usages",
            "total_memory_mb_usage",
            "total_vcpus_usage",
            "total_local_gb_usage",
        )
        column_headers = (
            "Project",
            "Servers",
            "RAM MB-Hours",
            "CPU Hours",
            "Disk GB-Hours",
        )

        data = utils.get_item_properties(
            usage, columns, formatters=_formatters(None)
        )
        return column_headers, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstackclient.identity.v2_0 import catalog
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class TestCatalog(utils.TestCommand):
    service_catalog = identity_fakes.FakeCatalog.create_catalog()

    def setUp(self):
        super().setUp()

        self.sc_mock = mock.Mock()
        self.sc_mock.service_catalog.catalog.return_value = [
            self.service_catalog,
        ]

        self.auth_mock = mock.Mock()
        self.app.client_manager.session = self.auth_mock

        self.auth_mock.auth.get_auth_ref.return_value = self.sc_mock


class TestCatalogList(TestCatalog):
    columns = (
        'Name',
        'Type',
        'Endpoints',
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = catalog.ListCatalog(self.app, None)

    def test_catalog_list(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN,
            fake_service=self.service_catalog,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                'supernova',
                'compute',
                catalog.EndpointsColumn(
                    auth_ref.service_catalog.catalog[0]['endpoints']
                ),
            ),
        )
        self.assertCountEqual(datalist, tuple(data))

    def test_catalog_list_with_endpoint_url(self):
        attr = {
            'id': 'qwertyuiop',
            'type': 'compute',
            'name': 'supernova',
            'endpoints': [
                {
                    'region': 'one',
                    'publicURL': 'https://public.one.example.com',
                },
                {
                    'region': 'two',
                    'publicURL': 'https://public.two.example.com',
                    'internalURL': 'https://internal.two.example.com',
                },
            ],
        }
        service_catalog = identity_fakes.FakeCatalog.create_catalog(attr)
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN,
            fake_service=service_catalog,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                'supernova',
                'compute',
                catalog.EndpointsColumn(
                    auth_ref.service_catalog.catalog[0]['endpoints']
                ),
            ),
        )
        self.assertCountEqual(datalist, tuple(data))


class TestCatalogShow(TestCatalog):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = catalog.ShowCatalog(self.app, None)

    def test_catalog_show(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.UNSCOPED_TOKEN,
            fake_service=self.service_catalog,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = [
            'compute',
        ]
        verifylist = [
            ('service', 'compute'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = ('endpoints', 'id', 'name', 'type')
        self.assertEqual(collist, columns)
        datalist = (
            catalog.EndpointsColumn(
                auth_ref.service_catalog.catalog[0]['endpoints']
            ),
            self.service_catalog.id,
            'supernova',
            'compute',
        )
        self.assertCountEqual(datalist, data)


class TestFormatColumns(TestCatalog):
    def test_endpoints_column_human_readable(self):
        col = catalog.EndpointsColumn(self.service_catalog['endpoints'])
        self.assertEqual(
            'one\n  publicURL: https://public.one.example.com\n  '
            'internalURL: https://internal.one.example.com\n  '
            'adminURL: https://admin.one.example.com\n'
            'two\n  publicURL: https://public.two.example.com\n  '
            'internalURL: https://internal.two.example.com\n  '
            'adminURL: https://admin.two.example.com\n'
            '<none>\n  publicURL: https://public.none.example.com\n  '
            'internalURL: https://internal.none.example.com\n  '
            'adminURL: https://admin.none.example.com\n',
            col.human_readable(),
        )

    def test_endpoints_column_human_readable_with_partial_endpoint_urls(self):
        endpoints = [
            {
                'region': 'one',
                'publicURL': 'https://public.one.example.com',
            },
            {
                'region': 'two',
                'publicURL': 'https://public.two.example.com',
                'internalURL': 'https://internal.two.example.com',
            },
        ]
        col = catalog.EndpointsColumn(endpoints)
        self.assertEqual(
            'one\n  publicURL: https://public.one.example.com\n'
            'two\n  publicURL: https://public.two.example.com\n  '
            'internalURL: https://internal.two.example.com\n',
            col.human_readable(),
        )
